﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Text.RegularExpressions;

namespace ScanHeadsman
{
    class ScanHeadsman
    {
        static private bool addMS = true;
        private delegate void ExtractionMethod(IList<InstrumentMethod> methods, string rawfilename);
        static private ExtractionMethod extractMethods = null;

        static private void extractAsText(IList<InstrumentMethod> methods, string rawFilename)
        {
            //export all methods as text
            StreamWriter methodfile = null;
            string methodFilename = Path.ChangeExtension(rawFilename, "methods.txt");

            Console.WriteLine("Extracting method to {0}", methodFilename);

            try
            {
                methodfile = new StreamWriter(methodFilename, false, Encoding.Unicode);
            }
            catch
            {
                throw new IOException(String.Format("Can not create output file: {0}", methodFilename));
            }

            for (int i = 0; i < methods.Count; i++)
            {
                methodfile.WriteLine(String.Format("Method {0} is {1}\n\n{2}", i, methods[i].Name, methods[i].Text));
            }

            methodfile.Close();
        }

        static private void extractAsXML(IList<InstrumentMethod> methods, string rawFilename)
        {
            //extract MS method as XML
            string methodFilename = Path.ChangeExtension(rawFilename, "method.xml");

            Console.WriteLine("Extracting MS method to {0}", methodFilename);

            for (int i = 0; i < methods.Count; i++)
            {
                if (methods[i].Name == "TNG-Calcium") // only TNG is supported so far
                {
                    MSMethod.serialize(new MSMethod(methods[i].Text), methodFilename);
                    break;
                }
            }
        }

        static private void extractAsCSV(IList<InstrumentMethod> methods, string rawFilename)
        {
            //extract MS experiments as CSV
            string methodFilename = Path.ChangeExtension(rawFilename, "experiments.csv");

            Console.WriteLine("Extracting experiments from MS method to {0}", methodFilename);

            for (int i = 0; i < methods.Count; i++)
            {
                if (methods[i].Name == "TNG-Calcium") // only TNG is supported so far
                {
                    MSMethod method = new MSMethod(methods[i].Text);
                    ToCSV(method.prepareForCSV(), methodFilename);
                    break;
                }
            }
        }

        static void ToCSV(IList<Dictionary<string, string>> data, string outputFilename, string columnSplitter = ",", string nan = "NA")
        {
            List<string> headers = new List<string>();
            int index = 0;
            Dictionary<string, Dictionary<int, string>> columns = new Dictionary<string, Dictionary<int, string>>();
            StreamWriter outfile = null;

            foreach (Dictionary<string, string> item in data)
            {
                foreach (KeyValuePair<string, string> kvpair in item)
                {
                    if (!headers.Contains(kvpair.Key))
                    {
                        headers.Add(kvpair.Key);
                        columns.Add(kvpair.Key, new Dictionary<int, string>());
                    }

                    columns[kvpair.Key][index] = kvpair.Value;
                }

                index++;
            }

            try
            {
                outfile = new StreamWriter(outputFilename, false, Encoding.ASCII);
            }
            catch
            {
                throw new IOException(String.Format("Can not create output file: {0}", outputFilename));
            }

            outfile.WriteLine(String.Join(columnSplitter, headers));

            for (int i = 0; i < index; i++)
            {
                List<string> row = new List<string>();
                foreach (string key in headers)
                {
                    try
                    {
                        if (columns[key][i].Contains(columnSplitter)) columns[key][i] = String.Format("\"{0}\"", columns[key][i]);
                        row.Add(columns[key][i]);
                    }
                    catch (KeyNotFoundException)
                    {
                        row.Add(nan);
                    }
                }
                outfile.WriteLine(String.Join(columnSplitter, row));
            }

            outfile.Close();
        }

        static void processRawFile(string rawFilename, string outputFile)
        {
            //open raw file
            ThermoRawfileReader raw = new ThermoRawfileReader();
            try
            {
                Console.WriteLine("Processing of {0}", rawFilename);
                raw.Open(rawFilename);
            }
            catch
            {
                Console.WriteLine("Can't open raw file: {0}", rawFilename);
                return;
            }

            //extract method information
            if (extractMethods != null)
            {
                extractMethods(raw.GetMethods(), rawFilename);
            }

            List<Dictionary<string, string>> harvest = new List<Dictionary<string, string>>();

            Regex TargetExpr = new Regex(@"(?<=T\=)\S+");
            Regex UfillExpr = new Regex(@"(?<=Ufill\=)\S+");
            Regex HCDEExpr = new Regex(@"(?<=HCD\=)\d+(?=eV)");
            Regex LMExpr = new Regex(@"Lock\(.*(\d)\/\d.*?(\S+)ppm\)");

            for (int scan = raw.FirstScan; scan <= raw.LastScan; scan++)
            {
                Dictionary<string, string> head = new Dictionary<string, string>();
                head.Add("Scan", scan.ToString());
                head.Add("MSOrder", raw.GetMSOrder(scan).ToString());
                head.Add("FilterString", raw.GetFilterString(scan));
                head.Add("DataType", raw.IsCentroided(scan));

                if (int.Parse(head["MSOrder"]) > 1)
                {
                    IList<FragmentationEvent> events = raw.ExtractFragmentationData(head["FilterString"]);

                    for (int i = 0; i < events.Count; i++)
                    {
                        head[String.Format("SelectedMass{0}", i + 1)] = events[i].SelectedMass.ToString(".00");
                        head[String.Format("Activation{0}", i + 1)] = events[i].Activation;
                        head[String.Format("Energy{0}", i + 1)] = events[i].Energy.ToString(".00");
                    }

                }

                foreach (KeyValuePair<string, string> item in raw.GetScanEventParsed(scan))
                {
                    head[item.Key] = item.Value;
                }

                ScanHeader header = raw.GetScanHeader(scan);
                head["RT [min]"] = header.RT.ToString(".00000");
                head["LowMass"] = header.LowMass.ToString(".00000");
                head["HighMass"] = header.HighMass.ToString(".00000");
                head["TIC"] = header.TIC.ToString(".0");
                head["BasePeakPosition"] = header.BasePeakPosition.ToString(".00000");
                head["BasePeakIntensity"] = header.BasePeakIntensity.ToString(".0");


                foreach (KeyValuePair<string, string> item in raw.GetTrailerExtra(scan))
                {
                    head[item.Key] = item.Value;
                }

                //squize some information from older instruments
                if (head.ContainsKey("FT Analyzer Settings"))
                {
                    head["TargetAGC"] = TargetExpr.Match(head["FT Analyzer Settings"]).Value;
                }

                if (head.ContainsKey("FT Analyzer Message"))
                {
                    string ufillString = UfillExpr.Match(head["FT Analyzer Message"]).Value;
                    head["Underfill"] = (ufillString == "") ? "1.0" : ufillString;
                    string HCDEString = HCDEExpr.Match(head["FT Analyzer Message"]).Value;
                    head["HCD Energy [eV]"] = (HCDEString == "") ? "0.0" : HCDEString;
                    Match LMString = LMExpr.Match(head["FT Analyzer Message"]);
                    head["NrLockMasses"] = LMString.Groups[1].Value;
                    head["LockMassDeviation [ppm]"] = LMString.Groups[2].Value;
                }

                //Mass Data
                MSData msData = new MSData();
                if (head["Analyzer"] == "FTMS")//use high precision method for FTMS scans
                {
                    msData = raw.GetExtendedSpectrumData(scan);
                    if (msData.mzArray.Length == 0 && head["DataType"] == "Centroided") //no labels in the data but it is centroided
                    {
                        msData = raw.GetSpectrumData(scan, false, 0);
                    }
                }
                else
                {
                    msData = raw.GetSpectrumData(scan);
                }

                head.Add("NrCentroids", msData.mzArray.Length.ToString());

                if (addMS)
                {
                    head.Add("mzArray", String.Format("\"{0}\"", String.Join(" ", msData.mzArray.Select(x => x.ToString(".00000")))));
                    head.Add("intArray", String.Format("\"{0}\"", String.Join(" ", msData.intArray.Select(x => x.ToString(".0")))));
                }

                //fix some weird stuff from trialer info
                if(head["MSOrder"] == "1")
                {
                    head["Monoisotopic M/Z"] = "";
                    head["Charge State"] = "";
                }

                harvest.Add(head);

                if (harvest.Count % 100 == 0)
                {
                    Console.Write("\rProcessing scan {0} of {1}", scan, raw.LastScan);
                }
            }

            Console.WriteLine("\rScan processing done; Total: {0} scans\nWrinting to CSV file.", harvest.Count);
            ToCSV(harvest, outputFile);
            Console.WriteLine("Done");

        }

        static List<string> ParseCLI(string[] cliArgs)
        {
            List<string> arguments = new List<string>();

            foreach (string item in cliArgs)
            {
                if (!item.StartsWith("--"))
                {
                    arguments.Add(item);
                }
                else
                {
                    if (item.ToLower() == "--noms") addMS = false;
                    else if (item.ToLower().StartsWith("--methods"))
                    {
                        //selecting proper format
                        string methodFormat = item.ToLower().Split(new char[] { ':' }).ElementAtOrDefault(1);

                        switch (methodFormat)
                        {
                            case "xml": extractMethods = extractAsXML; break;
                            case "csv": extractMethods = extractAsCSV; break;
                            case "text": extractMethods = extractAsText; break;
                            default: extractMethods = extractAsText; break;
                        }

                    }
                    else Console.WriteLine("Ignoring unknown argument - {0}", item);
                }
            }

            return arguments;
        }

        static void Main(string[] args)
        {
            int numFiles = 0;

            List<string> cliArgs = ParseCLI(args);

            switch (cliArgs.Count)
            {
                case 0:
                    Console.WriteLine("No arguments provided\nLooking for raw files in current folder\nUsing default extension for result");
                    foreach (string rawfileName in Directory.GetFiles(Directory.GetCurrentDirectory(), "*.raw"))
                    {
                        processRawFile(rawfileName, Path.ChangeExtension(rawfileName, "txt"));
                        numFiles++;
                    }
                    break;
                case 1:
                    if (Directory.Exists(cliArgs[0]))
                    {
                        Console.WriteLine("The argument is a folder\nLooking for raw files in it\nUsing default extension for result");
                        foreach (string rawfileName in Directory.GetFiles(cliArgs[0], "*.raw"))
                        {
                            processRawFile(rawfileName, Path.ChangeExtension(rawfileName, "txt"));
                            numFiles++;
                        }
                    }
                    else if (File.Exists(cliArgs[0]))
                    {
                        Console.WriteLine("The argument is a filename\nUsing default output name: {0}", Path.ChangeExtension(cliArgs[0], "txt"));
                        processRawFile(cliArgs[0], Path.ChangeExtension(cliArgs[0], "txt"));
                        numFiles++;
                    }
                    else
                    {
                        throw new IOException(String.Format("Can not find path: {0}", cliArgs[0]));
                    }
                    break;
                default:
                    if (File.Exists(cliArgs[0]))
                    {
                        processRawFile(cliArgs[0], cliArgs[1]);
                    }
                    else
                    {
                        throw new IOException(String.Format("First argument should be valid filename; {0} received", cliArgs[0]));
                    }
                    numFiles++;
                    break;
            }
            Console.WriteLine("{0} files processed", numFiles);
        }
    }
    
}
