﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using MSFileReaderLib;
using System.Text.RegularExpressions;

namespace ScanHeadsman
{
    public struct MSData
    {
        public double[] mzArray; // m/z
        public double[] intArray; // intensity
        public double[] resArray; // resolution
        public double[] blArray; // baseline
        public double[] noiseArray; // noise
        public int[] zArray; // charge
        public string[] flagArray; // flags
    }

    public struct FragmentationEvent
    {
        public double SelectedMass;
        public string Activation;
        public double Energy;
    }

    public struct ScanHeader
    {
        public double RT;
        public double LowMass;
        public double HighMass;
        public double TIC;
        public double BasePeakPosition;
        public double BasePeakIntensity;
    }

    public struct InstrumentMethod
    {
        public string Name;
        public string Text;
    }

    public class ThermoRawfileReader : Object
    {
        private IXRawfile5 rawfile = (IXRawfile5)new MSFileReader_XRawfile();
        private int nrMS = -1;
        private int firstScan = -1;
        private int lastScan = -1;

        private Regex fEventExpr = new Regex(@"\S+@\S+");
        private Regex activationExpr = new Regex(@"(\D+)([0-9.]+)");

        public int FirstScan
        {
            get
            {
                return this.firstScan;
            }
        }

        public int LastScan
        {
            get
            {
                return this.lastScan;
            }
        }

        public int MSCount
        {
            get
            {
                return this.nrMS;
            }
        }

        public ThermoRawfileReader()
        {
            //empty constructor
        }

        public ThermoRawfileReader(string filePath)
        {
            this.Open(filePath);
        }

        public void Open(string filePath)
        {
            rawfile.Open(filePath);

            rawfile.GetNumberOfControllersOfType(0, ref nrMS); // 0 - refers to MS devices
            rawfile.SetCurrentController(0, 1); //set the first MS controller as active
            rawfile.GetFirstSpectrumNumber(ref firstScan);
            rawfile.GetLastSpectrumNumber(ref lastScan);
        }

        public void Close()
        {
            rawfile.Close();
        }

        public void SetController(int controller)
        {
            if (controller <= nrMS) rawfile.SetCurrentController(0, controller);
        }

        public Dictionary<string, string> GetTrailerExtra(int scan)
        {
            int trailerSize = -1;
            object trailerLabels = null;
            object trailerValues = null;
            Dictionary<string, string> trailerData = new Dictionary<string, string>();

            rawfile.GetTrailerExtraForScanNum(scan, ref trailerLabels, ref trailerValues, ref trailerSize);

            //cast as array of stings
            string[] labels = (string[])trailerLabels;
            string[] values = (string[])trailerValues;

            for (int i = 0; i < trailerSize; i++)
            {
                //ignore empty labels
                if (labels[i].Trim().Length > 0) trailerData[labels[i].Substring(0, labels[i].Length - 1)] = values[i];
            }

            return trailerData;
        }

        public MSData GetFTMSSpectrumData(int scan) //mark for removal
        {
            MSData spectrumData = new MSData();

            object data = null;
            int listSize = -1;

            //This method returns FTMS data (surprize!)
            rawfile.GetMassPrecisionEstimate(scan, ref data, ref listSize);

            //cast to array
            double[,] massData = (double[,])data;

            //save the data
            spectrumData.mzArray = new double[listSize];
            spectrumData.intArray = new double[listSize];
            spectrumData.resArray = new double[listSize];
            Buffer.BlockCopy(massData, 0, spectrumData.intArray, 0, listSize * sizeof(double));
            Buffer.BlockCopy(massData, listSize * sizeof(double), spectrumData.mzArray, 0, listSize * sizeof(double));
            Buffer.BlockCopy(massData, 4 * listSize * sizeof(double), spectrumData.resArray, 0, listSize * sizeof(double));

            return spectrumData;
        }

        public MSData GetSpectrumData(int scan, bool centroided = true, double centroidingWidth = 0.1)
        {
            MSData spectrumData = new MSData();

            object data = null;
            object flags = null;
            int listSize = -1;

            int c = centroided ? 1 : 0;

            //(ScanNr, ScanFilter, IntensityCuttoffType, IntensityCuttoffValue, MaxNumberOfPeaks, Centroided, CentroidingWidth, MassList, flagList, SizeOfList)
            //ScanFilter: null - no filter
            //IntensityCuttoffType: 0 - No cuttoff
            //MaxNumberOfPeaks: 0 - all peaks
            //Centroided: (bool flag) 0 - profile, 1 - centroided
            //CentroidingWidth: window width used in centroiding
            rawfile.GetMassListFromScanNum(ref scan, null, 0, 0, 0, c, ref centroidingWidth, ref data, ref flags, ref listSize);

            //cast to array
            double[,] massData = (double[,])data;

            //save the data
            spectrumData.mzArray = new double[listSize];
            spectrumData.intArray = new double[listSize];
            Buffer.BlockCopy(massData, 0, spectrumData.mzArray, 0, listSize * sizeof(double));
            Buffer.BlockCopy(massData, listSize * sizeof(double), spectrumData.intArray, 0, listSize * sizeof(double));

            return spectrumData;
        }

        public MSData GetExtendedSpectrumData(int scan)
        {
            MSData spectrumData = new MSData();

            object data = null;
            object flags = null;
            
            //(LabelData, flags, Scan)
            rawfile.GetLabelData(ref data, ref flags, ref scan);

            double[,] msData = (double[,])data;
            byte[,] flagData = (byte[,])flags;

            int length =  msData.GetLength(1); // get length of the array

            //save the data
            spectrumData.mzArray = new double[length];
            spectrumData.intArray = new double[length];
            spectrumData.resArray = new double[length];
            spectrumData.blArray = new double[length];
            spectrumData.noiseArray = new double[length];
            spectrumData.zArray = new int[length];
            spectrumData.flagArray = new string[length];
            
            Buffer.BlockCopy(msData, 0, spectrumData.mzArray, 0, length * sizeof(double));
            Buffer.BlockCopy(msData, length * sizeof(double), spectrumData.intArray, 0, length * sizeof(double));
            Buffer.BlockCopy(msData, 2 * length * sizeof(double), spectrumData.resArray, 0, length * sizeof(double));
            Buffer.BlockCopy(msData, 3 * length * sizeof(double), spectrumData.blArray, 0, length * sizeof(double));
            Buffer.BlockCopy(msData, 4 * length * sizeof(double), spectrumData.noiseArray, 0, length * sizeof(double));

            for (int i = 0; i < length; i++)
            {
                spectrumData.zArray[i] = (int)msData[5, i];
                spectrumData.flagArray[i] = 
                    String.Format("{0}{1}{2}{3}{4}{5}",
                    (flagData[0, i] != 0) ? "S": "", //Saturated
                    (flagData[1, i] != 0) ? "F" : "", //Fragmented
                    (flagData[2, i] != 0) ? "M" : "", //Merged
                    (flagData[3, i] != 0) ? "E" : "", //Exception
                    (flagData[4, i] != 0) ? "R" : "", //Reference
                    (flagData[5, i] != 0) ? "m" : ""); //Modified
            }

            return spectrumData;
        }

        public string GetScanEventString(int scan)
        {
            string scanEvent = null;

            rawfile.GetScanEventForScanNum(scan, ref scanEvent);

            return scanEvent;
        }

        public Dictionary<string,string> GetScanEventParsed(int scan)
        {
            string scanEvent = null;

            rawfile.GetScanEventForScanNum(scan, ref scanEvent);

            return this.ParseScanEvent(scanEvent);
        }

        public string GetFilterString(int scan)
        {
            string filterString = null;

            rawfile.GetFilterForScanNum(scan, ref filterString);

            return filterString;
        }

        public ScanHeader GetScanHeader(int scan)
        {
            ScanHeader scanHeader = new ScanHeader();

            int numPoints = -1; //number of points in scan
            double startTime = -1; //RT
            double lowMass = -1; //lowest observed mass
            double highMass = -1; //highest observed mass
            double tic = -1; //TIC
            double baseMass = -1; //Basepeak position
            double baseIntensity = -1; //Basepeak intensity
            int numChannels = -1; //number of channels (not set for MS)
            int uniformTime = -1; //Uniform sampling (not set for MS)
            double frequency = -1; //Frequency (not set for MS)

            rawfile.GetScanHeaderInfoForScanNum(scan, ref numPoints, ref startTime, ref lowMass, ref highMass,
                ref tic, ref baseMass, ref baseIntensity, ref numChannels, ref uniformTime, ref frequency);

            scanHeader.RT = startTime;
            scanHeader.LowMass = lowMass;
            scanHeader.HighMass = highMass;
            scanHeader.TIC = tic;
            scanHeader.BasePeakPosition = baseMass;
            scanHeader.BasePeakIntensity = baseIntensity;
            
            return scanHeader;
        }

        public int GetMSOrder(int scan)
        {
            int order = -1;

            rawfile.GetMSOrderForScanNum(scan, ref order);

            return order;
        }

        public List<InstrumentMethod> GetMethods()
        {
            int numMethods = -1;
            object methodNames = null;
            rawfile.GetInstMethodNames(ref numMethods, ref methodNames);

            string[] methodIDs = (string[])methodNames;

            List<InstrumentMethod> methods = new List<InstrumentMethod>(numMethods);

            for (int n = 0; n < numMethods; n++)
            {
                string methodText = null;
                rawfile.GetInstMethod(n, ref methodText);
                methods.Add(new InstrumentMethod{Name = methodIDs[n], Text = methodText});
            }

            return methods;
        }

        public object GetPrecursorInfo(int scan)
        {
            int size = -1;
            object precursorInfo = null;

            try
            {
                rawfile.GetPrecursorInfoFromScanNum(scan, ref precursorInfo, ref size);
            }
            catch
            {
                return null;
            }

            return precursorInfo;
        }

        public List<string> GetFilters()
        {
            object filters = null;
            int size = -1;
            rawfile.GetFilters(ref filters, ref size);

            return new List<string>((string[])filters);
        }

        public string IsCentroided(int scan)
        {
            int c = -1;
            rawfile.IsCentroidScanForScanNum(scan, ref c);

            switch(c)
            {
                case 0: return "Profile";
                case 1: return "Centroided";
                default: return "Unknown";
            }
        }

        public IList<FragmentationEvent> ExtractFragmentationData(string scanFilter)
        {
            List<FragmentationEvent> fragmentationData = new List<FragmentationEvent>();

            foreach (Match group in fEventExpr.Matches(scanFilter))
                //group is something like <number>@<letters><number>(the part with @ can repeat many times)
            {
                string[] parts = group.Value.Split('@');
                FragmentationEvent fEvent = new FragmentationEvent();
                fEvent.SelectedMass = Double.Parse(parts[0]);//first part is selected mass
                for (int i = 1; i < parts.Length; i++)//second part and further is activations
                {
                    Match m = activationExpr.Match(parts[i]);//each part is something like hcd5.00 
                    fEvent.Activation = m.Groups[1].Value.ToUpper();
                    fEvent.Energy = Double.Parse(m.Groups[2].Value);
                    fragmentationData.Add(fEvent);
                    fEvent = new FragmentationEvent();//if we have multiple fragmentation add new event with selected mass 0
                    fEvent.SelectedMass = 0;
                }
                
            }

            return fragmentationData;
        }

        private Dictionary<string, string> ParseScanEvent (string scanEvent)
        {
            Dictionary<string, string> parsedEvent = new Dictionary<string, string>();

            foreach (string item in scanEvent.Split())
            {
                switch (item)
                {
                    case "+": parsedEvent["Polarity"] = "Positive"; break;
                    case "-": parsedEvent["Polarity"] = "Negative"; break;
                    case "Full": parsedEvent["ScanType"] = "Full"; break;
                    case "Z": parsedEvent["ScanType"] = "Zoom"; break;
                    case "SIM": parsedEvent["ScanType"] = "SIM"; break;
                    case "SRM": parsedEvent["ScanType"] = "SRM"; break;
                    case "CRM": parsedEvent["ScanType"] = "CRM"; break;
                    case "Q1MS": parsedEvent["ScanType"] = "Q1MS"; break;
                    case "Q3MS": parsedEvent["ScanType"] = "Q3MS"; break;
                    case "ITMS": parsedEvent["Analyzer"] = "ITMS"; break;
                    case "FTMS": parsedEvent["Analyzer"] = "FTMS"; break;
                    case "TQMS": parsedEvent["Analyzer"] = "TQMS"; break;
                    case "SQMS": parsedEvent["Analyzer"] = "SQMS"; break;
                    case "TOFMS": parsedEvent["Analyzer"] = "TOFMS"; break;
                    case "Sector": parsedEvent["Analyzer"] = "Sector"; break;
                    case "APCI": parsedEvent["Ionization"] = "APCI"; break;
                    case "ESI": parsedEvent["Ionization"] = "ESI"; break;
                    case "EI": parsedEvent["Ionization"] = "EI"; break;
                    case "CI": parsedEvent["Ionization"] = "CI"; break;
                    case "NSI": parsedEvent["Ionization"] = "NSI"; break;
                    case "FAB": parsedEvent["Ionization"] = "FAB"; break;
                    case "TSP": parsedEvent["Ionization"] = "TSP"; break;
                    case "FD": parsedEvent["Ionization"] = "FD"; break;
                    case "MALDI": parsedEvent["Ionization"] = "MALDI"; break;
                    case "GD": parsedEvent["Ionization"] = "GD"; break;
                    default:
                        {
                            if (item.StartsWith("!"))
                            {
                                try
                                {
                                parsedEvent[ScanEventCode[item.Substring(1)]] = "Off";
                                }
                                catch
                                {
                                    //nothing
                                }
                            }
                            else
                            {
                                try
                                {
                                    parsedEvent[ScanEventCode[item]] = "On";
                                }
                                catch
                                {
                                    //nothing
                                }
                            }
                            break;
                        }
                }
                
            }

            return parsedEvent;
        }

        private Dictionary<string, string> ScanEventCode = new Dictionary<string,string>
        {
            {"d", "DependentScan"},
            {"t", "TurboScan"},
            {"sid", "Source CID"},
            {"pi", "PhotoIonization"},
            {"cv", "CompensationVoltage"},
            {"det", "DetectorValid"},
            {"E", "Enhanced"},
            {"w", "Wideband"},
            {"sa", "SupplementalActivation"},
            {"msa", "MultistateActivation"},
            {"corona", "Corona"},
            {"AM", "AccurateMass"},
            {"AMI", "AccurateMass"},
            {"u", "UltraScan"},
            {"BSCAN", "BSectorScan"},
            {"ESCAN", "ESectorScan"},
            {"lock", "LockMass"},
            {"msx", "Multiplex"},
            {"nptr", "NPTR"}
        };
    }
}
